!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CCC
!CCC Name       : SUBROUTINE CALRATE 
!CCC Purpose    : To compute thermal rates (or yields) for each product channel
!CCC              as functions of both T and P at a particular angular momentum
!CCC 
!CCC Authors    : Thanh Lam Nguyen (thanhlamnguyen08@gmail.com)
!CCC              John F. Stanton  (jfstanton@gmail.com)
!CCC
!CCC Date       : Apr. 18, 2014
!CCC Where      : The University of Texas at Austin
!CCC
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


       SUBROUTINE CALRATE_ASYM(Jmin,Jmax,MAXJ, NJ,DELE,Maxi,FE,NN,NTS,Np,IND,Edown,T,Nsink,suJ,rKp,RKc,&       
                ESOLVER, BB, BBT, TSS, TDS, CNAME, CNM, CNP, IL, KL, SYMRE, SYMTS,&
                FEDISS, NDISS, NDI, NDJ, CDSTS, Rate_inf, NsinkTS, ESP,&
                Tim0, TEnd, MM, MODELJ, MODELR, RKcc, Ev, MLIN, RKinf, MODELDA,&
                SE9, NTC, SNN, SNP, SRP, STime)


        IMPLICIT NONE

!!!  Jmin, Jmax  : Input angular momentum number; Jmin = Jmax in this case. 
!!!  MAXJ        : maximum angular momentum number
!!!  NJ          : step size of angular momentum number, J
!!!  DELE        : energy bin (grain energy)
!!!  Maxi        : maximum bin in energy-grained master equation (EGME)
!!!  FE          : initial energy distribution function, F(E,J)
!!!  NN          : a number of minima
!!!  NTS         : a number of TSs
!!!  Np          : a number of products
!!!  IND         : index of the lowest-lying intermediate / adduct
!!!  Edown       : down energy transfer (cm-1)
!!!  T           : temperature (K)
!!!  Nsink       : a vector of sinks that is set to collect stabilized
!!!                populations of intermediates 
!!!  suJ         : sum of F(E) at a particular J    
!!!  rKp         : collision frequency (1/s) at a considered pressure
!!!  RKc         : overall bi-molecular reaction thermal rate (cm3/molecule/s)
!!!                or set RKc = 1 to gain product yields. 
!!!  ESOLVER     : either "DENSM" or "SPASM"
!!!  BB          : vector of three rot. constants for minima
!!!  BBT         : vector of three rot. constants for TSs
!!!  TSS         : number of states for TSs
!!!  TDS         : number of states for minima
!!!  CNAME       : names of stationary points
!!!  CNM         : names of minima
!!!  CNP         : names of products
!!!  IL          : index of reactant
!!!  KL          : index of product(s)
!!!  Sigf        : forward reaction path degeneracy
!!!  Sigr        : reverse reaction path degeneracy
!!!  FEDISS      : "YES" or "NO" : to print out F(E)J at the moment of
!!!                     dissociation
!!!  NDISS       : number of F(E)J functions need to be printed:
!!!                from NDI leading to NDJ
!!!  NDI         : index of reactant/adduct
!!!  NDJ         : index of products

        Include 'declare.inc'

        Integer MAXJ, NJ, IERR, MM, MMM, Np, NN, Maxi
        Integer N, M, NTS, IND, INDN, INDM, lam
        Integer  I, J, Jev, II, JJ, K, L, NE, NETS, JJT, KK, LL
        Integer Jmin, Jmax, NsinkTS0
        Integer  Nsink(No), ip(No), IL(No), KL(No), NsinkTS(No)
        Integer indx(MM), NN0(NN,Np)

        Integer*8 I4, J4, M4

        Double precision SYMRE(No), SYMTS(No), Ev, RKinf, sigma

        Double precision TSS(NTS,Maxi), TDS(NN,Maxi)
        Double precision, Allocatable :: SS(:,:), DS(:,:)

        Double precision, Allocatable :: A(:,:), AA(:,:), AM(:)
        Double precision, Allocatable :: A0(:,:) 
        Double precision, Allocatable :: A1(:,:)
        Double precision, Allocatable :: A2(:,:), A2m(:,:)
        Double precision, Allocatable :: A3(:,:)

        Double precision, Allocatable :: rktp(:,:,:)
        Double precision, Allocatable :: rk(:,:,:)
        Double precision, Allocatable :: rp(:,:,:)
        Double precision, Allocatable :: P(:,:,:)

        Double precision FE(Maxi,0:MAXJ)

        Double precision b(NN*Maxi), b2(NN*Maxi), SST(NN*Maxi)
        Double precision b0(NN*Maxi), bbb(NN*Maxi)
        Double precision Rate(NN+Np), RateP(NN+Np)
        Double precision FF(Maxi), Fint(Maxi), RRKM(Maxi)

        Double precision Sigf(No), Sigr(No)
        Double precision S(NN,Maxi), EER(MM), WR(MM), WI(MM)
        Double precision VR(MM,MM), VL(MM,MM)
        Double precision VRm(MM,MM)

        Double precision T1(Maxi), AT0(Maxi)
        Double precision SS1(Maxi), DS1(Maxi), DUMMY(1,1)

        Double precision PP, beta, T, Tim, Tim0, TEnd, rKp, RKc(50), RateI
        Double precision E0, Edown, DELE, tp, tpp, ESP
        Double precision suJJ, su, suJ, suk, suf1, suf2, suJJJ
        Double precision su2, su3, su4, ENG, ssu(MM)
        Double precision BT(3), suE(NN+100), tpA(MM), tp0(NN,Np)

        Double precision CC(MM), CC1(MM), CC2(MM), ltp(MM)
        Double precision CC0(MM), CCI(MM), CON1

        Double precision RAT, Rate0(NN,NN+Np), RKcc

        Integer LWORK
        Double precision WORK(70*NN*MM)

        INTEGER IPIV(MM), LDA, LDB, INFO
        INTEGER NNN, N1, M1, IT, NAEJ


        INTEGER NTC, III
        Double precision SE9(9), Tim00
        Double precision SNN(NN,NTC), SNP(NN,NN+Np,NTC)
        Double precision SRP(NN,NN+Np,NTC) 
        Double precision STime(NTC)


        CHARACTER*20 CNAME(NM,No)
        CHARACTER*20 CNM(No), CNP(No)
        CHARACTER*5 ESOLVER
        CHARACTER*3 FEDISS
        Integer NDISS, NDI(NM), NDJ(No)
        CHARACTER*12 CDSTS(No)
        Character*8 MODELJ, MLIN(50)
        Character*6 MODELR
        Character*6 MODELDA

        EXTERNAL DGEEV, indexx

        CC(1:MM) = 0.0d0
        CC0(1:MM) = 0.0d0
        CCI(1:MM) = 0.0d0


!!!CCCCCCCCCCCCCCCC
        NNN = 0
        DO I=1, NN
                DO J=1, Np
                        NNN = NNN + 1
                        NN0(I,J) = NNN + NN
                ENDDO        
        ENDDO
!!!CCCCCCCCCCCCCCCC

        NsinkTS0 = NsinkTS(1)

        ALLOCATE ( SS(NTS,Maxi), DS(NN,Maxi) )

        DO I=1, Maxi
          AT0(I)=0.0d0
        ENDDO

        beta = 1.0d0/RT/T
 
        IF(Jmin.EQ.Jmax) THEN
                JJT = Jmin
                su=0.0D0
                su2 = 0.0d0
                Do I=1, Maxi
                        su=su+FE(I,JJT)
                        su2 = su2 + (I-1)*DELE*FE(I,JJT)
                Enddo
                suJJ = su
        ELSE
                suJJ = suJ
        ENDIF

        suJJJ = suJJ

!!!        write(*,995) "Sum of F(E,J) = ", su
!!!        write(666,994) su

        su2 =0.0d0

        DO I=1, Maxi
                b(I) = FE(I,JJT)/su
                su2 = su2 + b(I)
                CC0(I) = b(I)
        ENDDO

!!!!!!!!

        DO K=2, NN
        DO I=1, Maxi
                CC0(I+(K-1)*(Maxi+Np)) = 0.0d0
        ENDDO
        ENDDO

!!!!!!!
 
!!!        write(*,994) su2

994     FORMAT(ES20.10E3)
995     FORMAT(A15,2X,ES20.10E3)

        IF(suJJ.eq.0.0d0) GOTO 1111

!CCCCCCCCCCCCCCCCCCCCCCCCCCCC

        DO I=1, NN
          DO J=1, NN + Np 
           Rate0(I,J)=0.0d0   !!! initial reaction rates for each channel
          Enddo
        ENDDO

!!!       IF(suJJ.eq.0.0d0) goto 450   !!! Do not compute when sum of F(E)J = 0
!!!        DO JJT=Jmin, Jmax, NJ  !!! To start J loop

        JJT = Jmin  !!! Note that Jmin = Jmax


!!CCCCCCCCCCCCCCCCCCCCCCCCC

!!! For minima

        DO K=1, NN
          DO I=1, Maxi
                T1(I) = TDS(K,I)
          ENDDO
          DO I=1, 3
                BT(I) = BB(K,I)
          ENDDO
          CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT,MODELJ,MLIN(K)) !!! To convolute E,J
          DO I=1, Maxi
                DS(K,I)=DS1(I)/SYMRE(K)
          ENDDO
        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCCCC

!!! For TSs

        DO K=1, NTS

        IF(CDSTS(K).NE.'VRRKM_HO') THEN

        DO I=1, Maxi
                T1(I) = TSS(K,I)
        ENDDO
        DO I=1, 3
                BT(I) = BBT(K,I)
        ENDDO
        CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT,MODELJ,MLIN(K+NN))  !!! To convolute E,J
        DO I=1, Maxi
                SS(K,I)=SS1(I)/SYMTS(K)
        ENDDO

        ELSE
                SS(K,1) = TSS(K,1)/SYMTS(K)
                DO J=2, Maxi
                        SS(K,J) = SS(K,J-1) + TSS(K,J)/SYMTS(K)
                ENDDO
        ENDIF

        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE ( rk(NN,NN+Np,Maxi) )


        DO I=1, NN
        DO J=1, NN + Np
                DO K=1, Maxi
                        rk(I,J,K)=0.0d0  !!! Initial microcanonical rates for
!!!                                             each reaction channel
                ENDDO
        ENDDO
        ENDDO                

        DO L=1, NTS
        IF((IL(L).GT.0).and.(KL(L).GT.0)) THEN
            DO J=1, Maxi
                IF(DS(IL(L),J).GT.0.0d0) THEN                
  rk(IL(L),KL(L),J) = rk(IL(L),KL(L),J) + SS(L,J)/DS(IL(L),J)/hh
                ENDIF
                IF(KL(L).LE.NN) THEN
                  IF(DS(KL(L),J).GT.0.0d0) THEN 
  rk(KL(L),IL(L),J) = rk(KL(L),IL(L),J) + SS(L,J)/DS(KL(L),J)/hh
                  ENDIF
                ENDIF
            ENDDO
        ENDIF
        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCC
!!! For an empty grain
!!CCCCCCCCCCCCCCCCCCCCCC

        DO K=1, Maxi 
        DO I=1, NN 
        DO J=1, NN 
                tp = rk(I,J,K)*rk(J,I,K)
                IF(tp.eq.0.0d0) THEN
                        rk(I,J,K)=0.0d0
                        rk(J,I,K)=0.0d0
                ENDIF
        ENDDO
        ENDDO
        ENDDO

!!! END of an empty grain



        ALLOCATE (P(NN,Maxi,Maxi))

        DO IND = 1, NN
                DO I=1, Maxi
                        DS1(I)=DS(IND,I)
                ENDDO
	call CalP(P,DELE,Edown,Maxi,T,Nsink,NN,DS1,IND) !!! to compute energy transfer
!!!                                            probability function, P(E)J
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To compute P(E) and construct an A 
!C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To input all product channels from each WELL	
!C	
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rp(NN,NN+Np,Maxi))

        DO I=1, NN
                ip(I) = Np + 1 
        ENDDO

        DO I=1, NN
                DO J=1, Np + 1 
                IF(J.EQ.1) THEN
                        DO K=1, Maxi 
                                rp(I,J,K)=rKp
                        ENDDO
                ELSE
                        DO K=1, Maxi
                                rp(I,J,K)=rk(I,J+NN-1,K)
                        ENDDO
                ENDIF
                ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Input rate coefficients in kinetic scheme
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rktp(NN,NN+Np,Maxi))

        DO K=1, Maxi
        DO I=1, NN
                su = 0.0d0
                DO J=1, Np + NN
                IF(I.ne.J) THEN
                        su = su + rk(I,J,K)
                ENDIF
                ENDDO
                rktp(I,I,K) = su
        ENDDO
        ENDDO

        DO K=1, Maxi
        DO I=1, NN
                DO J=1, NN + Np 
                IF(I.ne.J) THEN
                        rktp(I,J,K) = rk(I,J,K)
                ENDIF
                ENDDO
        ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Computing a A-matrix
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE ( A(MM,MM) )
        ALLOCATE ( A2(MM,MM) )

        DO I=1, MM
        DO J=1, MM
                A(I,J) = 0.0d0
                A2(I,J) = 0.0d0
        ENDDO
        ENDDO

	DO KK=1, NN
		DO LL=1, NN
			CALL CalA(A,rKp,P,rktp,Maxi,KK,LL,NN,Np)
		ENDDO
	ENDDO

!!!CCCCCCCCCCCCCCCCCCCCCCCCC

        IF(Np.GT.0) THEN

        DO I = 1, NN
        DO J = 1, Np
        DO KK=1, Maxi
           A(J+Maxi*NN+(I-1)*Np, (I-1)*Maxi+KK) = rktp(I,J+NN,KK)
        ENDDO   !!! KK 
        ENDDO   !!! J
        ENDDO   !!! I
        ENDIF

!!CCCCCCCC

        IF(MODELR.EQ.'CAT2') THEN
        su4 = 0.0d0
        DO KK=1, Maxi*NN
                A(KK,Maxi*NN+1) = CC0(KK)*RKcc
                su4 = su4 - CC0(KK)*RKcc
        ENDDO

        A(Maxi*NN+1,Maxi*NN+1) = su4

        DO I=1, Maxi*NN
                CC0(I) = 0.0d0
        ENDDO
        CC0( Maxi*NN + 1 ) = 1.0d0
        ENDIF        

!!CCCCCCCC

        IF(Np.GT.0) THEN
        DO J=1, MM
                ssu(J) = 0.0D0
                DO I=1, MM
                        ssu(J) = ssu(J) + A(I,J)
                ENDDO
        ENDDO

        DO J=1, MM
                A(J,J) = A(J,J) - ssu(J)
        ENDDO
        ENDIF

!!CCCCCCCCCCCCCCCCCCCCCCCCC

        DEALLOCATE ( P )
 
993     FORMAT(I5,2X,ES20.10E3)
991     FORMAT(2(I5,2X),2(ES20.10E3,2X))
9991    FORMAT(ES20.12)
9992    FORMAT(2(ES20.10E3,2X))
99993    FORMAT(I5,2X,2(ES20.10E3,2X))


IF(MODELDA.EQ.'ARPACK') THEN

!!CCCCCCCCCCCCCCCCCCCCCCCCC

        KK = 0
        DO I=1, MM
                tp = A(I,I)
                ltp(I) = A(I,I)
                IF(tp.EQ.0.0d0) KK = KK + 1
        ENDDO

        MMM = MM - KK
        M4 = MM*MM

        ALLOCATE ( AM(M4) )
        I4 = 0
        DO I=1, MM
                DO J=1, MM
                tp = A(I,I)*A(J,J)
                IF(tp.NE.0.0d0) THEN
                        I4 = I4 + 1
                        AM(I4) = A(I,J)
                ENDIF
                ENDDO
        ENDDO

        DEALLOCATE ( A )
        ALLOCATE ( AA(MMM,MMM) )

        DO I=1, MMM
                DO J=1, MMM
                I4 = (I-1)*MMM + J
                AA(I,J) = AM(I4)
                ENDDO
        ENDDO

        DEALLOCATE ( AM )

!!CCCCCCCCCCCCCCCCCCCCCCCCC

!!!        sigma = -RKinf   !!! rate constant at the high-P limit

        sigma = -RKinf
        IF(MODELR.EQ.'CAT2') THEN
                CON1 = 1.0D+15
        ELSE
                CON1 = 1.0D0
        ENDIF
        CALL Call_ANEV(AA,MMM,sigma,CC1,CC2,CON1,SE9)

        DO I=1, 9
                SE9(I) = SE9(I)*suJJJ
        ENDDO
        DEALLOCATE ( AA )


!!!!!!!

ELSE IF(MODELDA.EQ.'LAPACK') THEN

        CALL DGEEV('N','V', MM, A, MM, WR, WI, DUMMY, 1, VR, MM, &
                WORK, 70*MM, IERR )

!!!        write(*,*) IERR
        IF(IERR.NE.0) STOP

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        CALL indexx(MM,WR,indx)

        DO I=1, MM
!!!                write(11112,9993) I, WR(indx(I))

                IF(abs(WR(I)).LT.Ev) WR(I)=0.0d0
                IF(abs(WR(indx(I))).LT.Ev) WR(indx(I))=0.0d0

        ENDDO

        tpp = abs(WR(indx(1)))/1.0D+15  !!! control low-P
        tp = rKp/1.0D+15   !!! control high-P

        IF(tpp.LT.tp) tpp=tp

        DO I=1, MM
                IF(WR(indx(I)).GT.0.0d0) WR(indx(I))=0.0D0
                tp = abs(WR(indx(I)))
                IF(tp.LE.tpp) WR(indx(I)) = 0.0D0
!!!                write(11113,9993) I, WR(indx(I))
        ENDDO

        Jev = 1
        RAT = WR(indx(Jev))
        DO I = 2, MM
                tp = WR(indx(I))
                IF((RAT.LT.tp).and.(tp.LT.0.0d0)) THEN
                        IF(dabs(tp).GE.ESP) THEN
                                RAT=tp
                                Jev=I
                        ENDIF
                ENDIF
        ENDDO

!!!        write(*,*) WR(indx(Jev))

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


        LDA = MM

        DO J=1, MM
        DO I=1, MM
                IF(I.EQ.J) THEN
                        VRm(I,J)=1.0d0
                ELSE
                        VRm(I,J)=0.0d0
                ENDIF
                A2(I,J) = VR(I,J)
        ENDDO
        ENDDO

        CALL dgesv(MM, MM, A2, LDA, IPIV, VRm, MM, INFO)

        DEALLOCATE ( A2 )

!!!        write(*,*) INFO
        IF(INFO.NE.0) STOP
!!!        write(*,*)


!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        DO J=1, MM
                CCI(J) = 0.0d0 
                DO K=1, MM
                        CCI(J) = CCI(J) + VRm(indx(J),K)*CC0(K)
                ENDDO
         ENDDO

!!CCCCCCCCCCCCCCCCCCCCCC
        DO I=1, NN
        DO J=1, Np
                tp0(I,J) = 0.0d0
        ENDDO
        ENDDO
!!CCCCCCCCCCCCCCCCCCCCCC

        Tim00 = Tim0

        II = 1
        Tim = Tim00*II
        III = 1

        STime(III) = Tim*suJJJ

        DO WHILE (Tim.LE.TEnd)

        DO I=1, MM
        CC(I) = 0.0d0
        DO J=1, MM
                IF(WR(indx(J)).LE.0.0d0) THEN
                tp = abs(WR(indx(J))*Tim)
                IF(tp.LT.60.0d0) THEN
                CC(I) = CC(I) + VR(I,indx(J))*exp(WR(indx(J))*Tim)*CCI(J)
                ENDIF
                ENDIF
        ENDDO
        ENDDO

        DO I=1, NN
                tp=0.0d0
                DO J=1, Maxi
                tp = tp + CC(J+(I-1)*Maxi)
                ENDDO
                SNN(I,III) = tp*suJJJ

                DO J = 1, Np
                tp = CC( Maxi*NN + (I-1)*Np + J )
                IF(abs(tp).LE.1.0d0) THEN
                SNP(I,J+NN,III) = tp*suJJJ
                ELSE
                SNP(I,J+NN,III) = 0.0d0
                ENDIF 
                ENDDO
        ENDDO

!!!!!!!

        KK = 0
        DO I=1, NN
        DO JJ=1, NN + Np
        IF(I.NE.JJ) THEN
        su2 = 0.0d0
        su3 = 0.0d0
        KK = KK + 1
        DO K=1, Maxi
                su3 = su3 + CC(K+(I-1)*Maxi)
                su2 = su2 + rk(I,JJ,K)*CC(K+(I-1)*Maxi)
                ENG = (K-1)*DELE
        ENDDO
        IF(su3.NE.0.0d0) THEN
                IF((su2/su3).GT.0.0d0) THEN
                SRP(I,JJ,III) = (su2/su3)*suJJJ
                ELSE
                SRP(I,JJ,III) = 0.0d0
                ENDIF
        ELSE
                SRP(I,JJ,III) = 0.0d0
        ENDIF
        ENDIF
        ENDDO
        ENDDO

93 FORMAT (ES20.10E3,2X,2(I5,2X),ES20.10E3)

        IF(Tim.EQ.10*Tim00) Then
                Tim00 = Tim00*10
                II = 1
        ENDIF

        II = II + 1
        Tim = Tim00*II

        III = III + 1
        STime(III) = Tim*suJJJ

        ENDDO

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

!!!        write(*,*) "Nine Lowest Eigenvalues = " 
        DO I=1, 9 
                INDM = Jev + I - 8 
                INDN = indx(INDM) 
                SE9(I) = WR(INDN)*suJJJ
        ENDDO

ELSE
        write(*,*) "Sth is wrong with the selection of diagonalizing M"
ENDIF

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

1111        DEALLOCATE ( rktp, rk, SS, DS, rp )

        RETURN
	END

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

