!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CCC
!CCC Name       : SUBROUTINE CALRATE 
!CCC Purpose    : To compute thermal rates (or yields) for each product channel
!CCC              as functions of both T and P at a particular angular momentum
!CCC 
!CCC Authors    : Thanh Lam Nguyen (thanhlamnguyen08@gmail.com)
!CCC              John F. Stanton  (jfstanton@gmail.com)
!CCC
!CCC Date       : Apr. 18, 2014
!CCC Where      : The University of Texas at Austin
!CCC
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


       SUBROUTINE CALRATE_SYMM(Jmin,Jmax,MAXJ, NJ,DELE,Maxi,FE,NN,NTS,Np,IND,Edown,T,Nsink,suJ,rKp,RKc,&       
                ESOLVER, BB, BBT, TSS, TDS, CNAME, CNM, CNP, IL, KL, SYMRE, SYMTS, &
                FEDISS, NDISS, NDI, NDJ, CDSTS, Rate_inf, NsinkTS, ESP,&
                Tim0, TEnd, MM, MODELJ, MODELR, RKcc, Ev, MLIN )

        IMPLICIT NONE

!!!  Jmin, Jmax  : Input angular momentum number; Jmin = Jmax in this case. 
!!!  MAXJ        : maximum angular momentum number
!!!  NJ          : step size of angular momentum number, J
!!!  DELE        : energy bin (grain energy)
!!!  Maxi        : maximum bin in energy-grained master equation (EGME)
!!!  FE          : initial energy distribution function, F(E,J)
!!!  NN          : a number of minima
!!!  NTS         : a number of TSs
!!!  Np          : a number of products
!!!  IND         : index of the lowest-lying intermediate / adduct
!!!  Edown       : down energy transfer (cm-1)
!!!  T           : temperature (K)
!!!  Nsink       : a vector of sinks that is set to collect stabilized
!!!                populations of intermediates 
!!!  suJ         : sum of F(E) at a particular J    
!!!  rKp         : collision frequency (1/s) at a considered pressure
!!!  RKc         : overall bi-molecular reaction thermal rate (cm3/molecule/s)
!!!                or set RKc = 1 to gain product yields. 
!!!  ESOLVER     : either "DENSM" or "SPASM"
!!!  BB          : vector of three rot. constants for minima
!!!  BBT         : vector of three rot. constants for TSs
!!!  TSS         : number of states for TSs
!!!  TDS         : number of states for minima
!!!  CNAME       : names of stationary points
!!!  CNM         : names of minima
!!!  CNP         : names of products
!!!  IL          : index of reactant
!!!  KL          : index of product(s)
!!!  Sigf        : forward reaction path degeneracy
!!!  Sigr        : reverse reaction path degeneracy
!!!  FEDISS      : "YES" or "NO" : to print out F(E)J at the moment of
!!!                     dissociation
!!!  NDISS       : number of F(E)J functions need to be printed:
!!!                from NDI leading to NDJ
!!!  NDI         : index of reactant/adduct
!!!  NDJ         : index of products

        Include 'declare.inc'

        Integer MAXJ, NJ, IERR, MM, Np, NN, Maxi
        Integer N, M, NTS, IND, INDN, INDM, lam
        Integer  I, J, Jev, II, JJ, K, L, NE, NETS, JJT, KK, LL
        Integer Jmin, Jmax, NsinkTS0, KKK, LLL, N0
        Integer  Nsink(No), ip(No), IL(No), KL(No), NsinkTS(No)
        Integer indx(MM), NN0(NN,Np)

integer, parameter :: qp = selected_real_kind(15, 307)

real(kind=qp) :: TSS(NTS,Maxi), TDS(NN,Maxi), fv1(MM), fv2(MM)
real(kind=qp), Allocatable :: SS(:,:), DS(:,:), FEJ(:,:)

real(kind=qp) :: SYMRE(No), SYMTS(No), Ev

real(kind=qp), Allocatable :: A(:,:)
real(kind=qp), Allocatable :: A0(:,:) 
real(kind=qp), Allocatable :: A1(:,:)
real(kind=qp), Allocatable :: A2(:,:), A2m(:,:)
real(kind=qp), Allocatable :: BA(:,:)

real(kind=qp), Allocatable :: rktp(:,:,:)
real(kind=qp), Allocatable :: rk(:,:,:)
real(kind=qp), Allocatable :: rp(:,:,:)
real(kind=qp), Allocatable :: P(:,:,:)

real(kind=qp) :: FE(Maxi,0:MAXJ)

real(kind=qp) :: b(NN*Maxi), b2(NN*Maxi), SST(NN*Maxi)
real(kind=qp) :: b0(NN*Maxi), bbb(MM)
real(kind=qp) ::  Rate(NN+Np), RateP(NN+Np)
real(kind=qp) :: FF(Maxi), Fint(Maxi), RRKM(Maxi)

real(kind=qp) :: Sigf(No), Sigr(No)
real(kind=qp) :: S(NN,Maxi), EER(MM), WR(MM), WI(MM)
real(kind=qp) ::  VR(MM,MM), VL(MM,MM)
real(kind=qp) :: VRm(MM,MM)

real(kind=qp), Allocatable :: VR2(:,:), VR2m(:,:)


real(kind=qp) :: T1(Maxi), AT0(Maxi)
real(kind=qp) :: SS1(Maxi), DS1(Maxi), DUMMY(1,1)
real(kind=qp) :: PP, beta, T, Tim, Tim0, TEnd, rKp, RKc(50), RateI
real(kind=qp) :: Edown, DELE, tp, tpp, ESP
real(kind=qp) :: suJJ, su, suJ, suk, suf1, suf2
real(kind=qp) :: su2, su3, su4, ENG, ssu(MM)
real(kind=qp) :: BT(3), suE(NN+100), tpA(MM), tp0(NN,Np)
real(kind=qp) :: CC(MM)
real(kind=qp) :: CC0(MM), CCI(MM)
real(kind=qp) :: RAT, Rate0(NN,NN+Np), RKcc 
real(kind=qp), Allocatable :: SP(:), SM(:)

        Integer LWORK
real(kind=qp) :: WORK(70*NN*MM)

        INTEGER IPIV(MM), LDA, LDB, INFO
        INTEGER NNN, N1, M1, IT, NAEJ

        CHARACTER*20 CNAME(NM,No)
        CHARACTER*20 CNM(No), CNP(No)
        CHARACTER*5 ESOLVER
        CHARACTER*3 FEDISS
        Integer NDISS, NDI(NM), NDJ(No)
        CHARACTER*12 CDSTS(No)
        Character*8 MODELJ, MLIN(50)
        Character*6 MODELR

!!!CCCCCCCCCCCCCCCC
        NNN = 0
        DO I=1, NN
                DO J=1, Np
                        NNN = NNN + 1
                        NN0(I,J) = NNN + NN
                ENDDO        
        ENDDO
!!!CCCCCCCCCCCCCCCC

        NsinkTS0 = NsinkTS(1)

        ALLOCATE ( SS(NTS,Maxi), DS(NN,Maxi), FEJ(NN+Np,Maxi) )

        FEJ(1:NN+Np,1:Maxi) = 0.0d0


        DO I=1, Maxi
          AT0(I)=0.0D0
        ENDDO

        beta = 1.0D0/RT/T
 
        IF(Jmin.EQ.Jmax) THEN
                JJT = Jmin
                su=0.0D0
                su2 = 0.0D0
                Do I=1, Maxi
                        su=su+FE(I,JJT)
                        su2 = su2 + (I-1)*DELE*FE(I,JJT)
                Enddo
                suJJ = su
        ELSE
                suJJ = suJ
        ENDIF

        write(*,995) "Sum of F(E,J) = ", su

        write(666,994) su

        su2 =0.0d0

        DO I=1, Maxi
                b(I) = FE(I,JJT)/su
                su2 = su2 + b(I)
                CC0(I) = b(I)
        ENDDO

!!!!!!!!

        DO K=2, NN
        DO I=1, Maxi
                CC0(I+(K-1)*(Maxi+Np)) = 0.0D0
        ENDDO
        ENDDO

!!!!!!!
 
        write(*,994) su2

994     FORMAT(ES20.10E3)
995     FORMAT(A15,2X,ES20.10E3)

        IF(suJJ.eq.0.0D0) GOTO 1111

!CCCCCCCCCCCCCCCCCCCCCCCCCCCC

        DO I=1, NN
          DO J=1, NN + Np 
           Rate0(I,J)=0.0D0   !!! initial reaction rates for each channel
          Enddo
        ENDDO

!!!       IF(suJJ.eq.0.0d0) goto 450   !!! Do not compute when sum of F(E)J = 0
!!!        DO JJT=Jmin, Jmax, NJ  !!! To start J loop

        JJT = Jmin  !!! Note that Jmin = Jmax


!!CCCCCCCCCCCCCCCCCCCCCCCCC

!!! For minima

        DO K=1, NN
          DO I=1, Maxi
                T1(I) = TDS(K,I)
          ENDDO
          DO I=1, 3
                BT(I) = BB(K,I)
          ENDDO
          CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT,MODELJ,MLIN(K)) !!! To convolute E,J
          DO I=1, Maxi
                DS(K,I)=DS1(I)/SYMRE(K)
          ENDDO
        ENDDO
!!CCCCCCCCCCCCCCCCCCCCCCCC

        DO K=1, NN
        DO I=1, Maxi
                ENG = (I-1)*DELE
                FEJ(K,I) = DS(K,I)*exp(-ENG*beta)
        ENDDO
        ENDDO 

!!CCCCCCCCCCCCCCCCCCCCCCCC

!!! For TSs

        DO K=1, NTS

        IF(CDSTS(K).NE.'VRRKM_HO') THEN

        DO I=1, Maxi
                T1(I) = TSS(K,I)
        ENDDO
        DO I=1, 3
                BT(I) = BBT(K,I)
        ENDDO
        CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT,MODELJ,MLIN(K+NN))  !!! To convolute E,J
        DO I=1, Maxi
                SS(K,I)=SS1(I)/SYMTS(K)
        ENDDO

        ELSE
                SS(K,1) = TSS(K,1)/SYMTS(K)
                DO J=2, Maxi
                        SS(K,J) = SS(K,J-1) + TSS(K,J)/SYMTS(K)
                ENDDO
        ENDIF

        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE ( rk(NN,NN+Np,Maxi) )


        DO I=1, NN
        DO J=1, NN + Np
                DO K=1, Maxi
                        rk(I,J,K)=0.0D0  !!! Initial microcanonical rates for
!!!                                             each reaction channel
                ENDDO
        ENDDO
        ENDDO                

        DO L=1, NTS
        IF((IL(L).GT.0).and.(KL(L).GT.0)) THEN
            DO J=1, Maxi
                IF(DS(IL(L),J).GT.0.0D0) THEN                
  rk(IL(L),KL(L),J) = rk(IL(L),KL(L),J) + SS(L,J)/DS(IL(L),J)/hh
                ENDIF
                IF(KL(L).LE.NN) THEN
                  IF(DS(KL(L),J).GT.0.0D0) THEN 
  rk(KL(L),IL(L),J) = rk(KL(L),IL(L),J) + SS(L,J)/DS(KL(L),J)/hh
                  ENDIF
                ENDIF
            ENDDO
        ENDIF
        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCC
!!! For an empty grain
!!CCCCCCCCCCCCCCCCCCCCCC

        DO K=1, Maxi 
        DO I=1, NN 
        DO J=1, NN 
                tp = rk(I,J,K)*rk(J,I,K)
                IF(tp.eq.0.0d0) THEN
                        rk(I,J,K)=0.0d0
                        rk(J,I,K)=0.0d0
                ENDIF
        ENDDO
        ENDDO
        ENDDO

!!! END of an empty grain


        ALLOCATE (P(NN,Maxi,Maxi))

        DO K = 1, NN
                DO I=1, Maxi
                        DS1(I)=DS(K,I)
                ENDDO
	call CalP(P,DELE,Edown,Maxi,T,Nsink,NN,DS1,K) !!! to compute energy transfer
!!!                                            probability function, P(E)J
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To compute P(E) and construct an A 
!C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To input all product channels from each WELL	
!C	
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rp(NN,NN+Np,Maxi))

        DO I=1, NN
                ip(I) = Np + 1 
        ENDDO

        DO I=1, NN
                DO J=1, Np + 1 
                IF(J.EQ.1) THEN
                        DO K=1, Maxi 
                                rp(I,J,K)=rKp
                        ENDDO
                ELSE
                        DO K=1, Maxi
                                rp(I,J,K)=rk(I,J+NN-1,K)
                        ENDDO
                ENDIF
                ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Input rate coefficients in kinetic scheme
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rktp(NN,NN+Np,Maxi))

        DO K=1, Maxi
        DO I=1, NN
                su = 0.0D0
                DO J=1, Np + NN
                IF(I.ne.J) THEN
                        su = su + rk(I,J,K)
                ENDIF
                ENDDO
                rktp(I,I,K) = su
        ENDDO
        ENDDO

        DO K=1, Maxi
        DO I=1, NN
                DO J=1, NN + Np 
                IF(I.ne.J) THEN
                        rktp(I,J,K) = rk(I,J,K)
                ENDIF
                ENDDO
        ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Computing a A-matrix
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE ( A(MM,MM) )

        A(1:MM,1:MM) = 0.0D0

	DO KK=1, NN
		DO LL=1, NN
			CALL CalA(A,rKp,P,rktp,Maxi,KK,LL,NN,Np)
		ENDDO
	ENDDO

!!!CCCCCCCCCCCCCCCCCCCCCCCCC

        IF(Np.GT.0) THEN

        DO I = 1, NN
        DO J = 1, Np
        DO KK=1, Maxi
           A(J+Maxi*NN+(I-1)*Np, (I-1)*Maxi+KK) = rktp(I,J+NN,KK)
        ENDDO   !!! KK 
        ENDDO   !!! J
        ENDDO   !!! I
        ENDIF

!!CCCCCCCC

        IF(MODELR.EQ.'CAT2') THEN
        su4 = 0.0D0
        DO KK=1, Maxi*NN
                A(KK,Maxi*NN+1) = CC0(KK)*RKcc
                su4 = su4 - CC0(KK)*RKcc
        ENDDO

        A(Maxi*NN+1,Maxi*NN+1) = su4

        DO I=1, Maxi*NN
                CC0(I) = 0.0D0
        ENDDO
        CC0( Maxi*NN + 1 ) = 1.0D0
        ENDIF        

!!CCCCCCCC

        ALLOCATE ( BA(MM,MM) )
        DO J=1, MM
        DO I=1, MM
                BA(I,J) = 0.0D0
        ENDDO
        ENDDO

!!====================

!!! To symmetrize an A matrix to a BA matrix

        ALLOCATE ( SP(MM), SM(MM) )

        DO I=1, MM
                SP(I) = 0.0d0
                SM(I) = 0.0d0
        ENDDO

        DO I=1, NN
        DO J=1, Maxi
                II = J + (I-1)*Maxi
                tp = FEJ(I,J)
                IF(tp.eq.0.0d0) THEN
                       SP(II) = 0.0d0
                ELSE 
                        SP(II) = 1.0D0/sqrt(FEJ(I,J))
                ENDIF
                SM(II) = sqrt(FEJ(I,J))
        ENDDO
        ENDDO

        DO I = 1, MM
        DO J = 1, MM
                BA(I,J) = SP(I)*A(I,J)*SM(J)
        ENDDO
        ENDDO


        DO I = 1, MM
        DO J = 1, I
                BA(I,J) = BA(J,I)
        ENDDO
        ENDDO


!!        DO I=1, Maxi*NN
!!        DO J=1, Maxi*NN
!!                write(9999,99099) BA(I,J)
!!        ENDDO
!!        ENDDO

99099   FORMAT (1(ES12.5,1x)) 

!!CCCCCCCCCCCCCCCCCCCCCCCCC

        DEALLOCATE ( P )
 
993     FORMAT(I5,2X,ES20.10E3)
991     FORMAT(2(I5,2X),2(ES20.10E3,2X))
9991    FORMAT(ES20.10E3)
9992    FORMAT(2(ES20.10E3,2X))
99993   FORMAT(I5,2X,2(ES20.10E3,2X))

        LWORK = -1
        CALL dsyev("V", "U", MM, BA, MM, WR, WORK, LWORK, IERR)
        LWORK = MIN(70*MM, INT(WORK(1)))
        CALL dsyev("V", "U", MM, BA, MM, WR, WORK, LWORK, IERR) 

        ALLOCATE ( VR2(MM,MM) )

        DO I=1, MM
        DO J=1, MM
                VR2(I,J) = BA(I,J)
        ENDDO
        ENDDO

        write(*,*) IERR, "Hello Hello"
        IF(IERR.NE.0) STOP

!!CCCCCCCCCCCCCCCCCCCC

        DO I=1, MM
                write(11114,9993) I, WR(I)
        ENDDO




!!! To delete all small (+ or -) eigenvalues, which are accurate within
!!! a double-precision floating-point.

        tpp = abs(WR(1))/1.0D+15  !!! control low-P
        tp = rKp/1.0D+13   !!! control high-P 

        IF(tpp.LT.tp) tpp=tp

        DO I=1, MM

                IF(abs(WR(I)).LT.Ev) WR(I) = 0.0d0

                IF(WR(I).GT.0.0d0) WR(I)=0.0D0
                tp = abs(WR(I))
                IF(tp.LE.tpp) WR(I) = 0.0D0
                write(11113,9993) I, WR(I)
        ENDDO

!!CCCCCCCCCCCCCCCCCCCC


!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


!!! To compute U^(-1) of a symmetric matrix BA

        ALLOCATE ( VR2m(MM,MM) )
        ALLOCATE ( A2(MM,MM) )

        LDA = MM
        INFO = 0
        write(*,*) INFO, "LIN_EQ_1"

        DO J=1, MM
        DO I=1, MM
                IF(I.EQ.J) THEN
                        VR2m(I,J)=1.0d0
                ELSE
                        VR2m(I,J)=0.0d0
                ENDIF
                A2(I,J) = VR2(I,J)
        ENDDO
        ENDDO

        CALL dgesv(MM, MM, A2, LDA, IPIV, VR2m, MM, INFO)

        write(*,*) INFO, "LIN_EQ_2"
        IF(INFO.NE.0) STOP
        DEALLOCATE ( A2 )

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

!!! To compute U and U^(-1) of an asymmetric (original) matrix A

        VR(1:MM,1:MM) = 0.0D0
        VRm(1:MM,1:MM) = 0.0D0

        DO I=1, MM
        DO J=1, MM
                VRm(I,J) = VR2m(I,J)*SP(J)
                VR(I,J) = SM(I)*VR2(I,J)
        ENDDO
        ENDDO

        DEALLOCATE ( VR2, VR2m )

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        Jev = 1 
        RAT = WR(Jev)
        DO I = 2, MM 
                tp = WR(I)
                IF((RAT.LT.tp).and.(tp.LT.0.0D0)) THEN
                        IF(abs(tp).GE.ESP/1.0D+06) THEN
                                RAT=tp
                                Jev=I
                        ENDIF
                ENDIF
        ENDDO

        write(*,*) WR(Jev)
        write(6661,*) WR(Jev)
        write(2001,9994) -WR(Jev)

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        DO J=1, MM
                CCI(J) = 0.0D0 
                DO K=1, MM
                        CCI(J) = CCI(J) + VRm(J,K)*CC0(K)
                ENDDO
         ENDDO

!!CCCCCCCCCCCCCCCCCCCCCC
        DO I=1, NN
        DO J=1, Np
                tp0(I,J) = 0.0D0
        ENDDO
        ENDDO
!!CCCCCCCCCCCCCCCCCCCCCC

        II = 1
        Tim = Tim0*II

        DO WHILE (Tim.LE.TEnd)

        DO I=1, MM
        CC(I) = 0.0D0

        DO J=1, MM
                CC(I) = CC(I) + VR(I,J)*exp(WR(J)*Tim)*CCI(J)
        ENDDO
        ENDDO

        DO I=1, NN
                tp=0.0D0
                DO J=1, Maxi
                tp = tp + CC(J+(I-1)*Maxi)
                ENDDO
                write(9000+I,9908) Tim, I, I, tp
                DO J = 1, Np
                tp = CC( Maxi*NN + (I-1)*Np + J )
        write(9000+NN0(I,J),9908) Tim, I, NN+J, tp 
                ENDDO
        ENDDO

9908    FORMAT(ES20.10E3,2X,I5,2X,I5,2X,ES20.10E3)

!!!!!!!

        KK = 0
        DO I=1, NN
        DO JJ=1, NN + Np
        IF(I.NE.JJ) THEN
        su2 = 0.0D0
        su3 = 0.0D0
        KK = KK + 1
        DO K=1, Maxi
                su3 = su3 + CC(K+(I-1)*Maxi)
                su2 = su2 + rk(I,JJ,K)*CC(K+(I-1)*Maxi)
                ENG = (K-1)*DELE
        ENDDO
        IF(su3.NE.0.0D0) THEN
                IF((su2/su3).GT.0.0D0) THEN
                write(8000+KK,93) Tim, I, JJ, (su2/su3)
                ELSE
                write(8000+KK,93) Tim, I, JJ, 0.0D0
                ENDIF
        ELSE
                write(8000+KK,93) Tim, I, JJ, 0.0D0
        ENDIF
        ENDIF
        ENDDO
        ENDDO

93 FORMAT (ES20.10E3,2X,2(I5,2X),ES20.10E3)

        IF(Tim.EQ.10*Tim0) Then
                Tim0 = Tim0*10
                II = 1
        ENDIF

        II = II + 1
        Tim = Tim0*II

        ENDDO

!!!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

!!!!!!

        write(*,*)
        write(*,*) "Nine Lowest Eigenvalues = " 
        DO I=1, 9
                INDM = Jev + I - 8
                INDN = INDM 
                write(*,'(I5,2X,ES15.5)') I, WR(INDN)
                write(6000+I,'(ES15.5)') WR(INDN)
        ENDDO

        write(*,*)
        write(*,*)

        write(*,*)
        write(*,*) "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"
        write(*,*)

        write(*,*)  "The second lowest eigenvalue"
        write(*,'(I5,2X,ES15.5)') Jev-1, WR(Jev-1)

        J=NNN + 1
        INDM = Jev-1
        INDN = INDM
        suE(J)=0.0D0
        DO I=1, MM
                suE(J)=suE(J) + VR(I,INDN)
        ENDDO

        DO I=1, MM
                VR(I,INDN) = VR(I,INDN)/suE(J)
        ENDDO

        KK = 0
        DO I=1, NN
        DO JJ=1, NN + Np
        IF(I.NE.JJ) THEN    
        su2 = 0.0D0
        su3 = 0.0D0
        KK = KK + 1 
        DO K=1, Maxi
                INDN = Jev-1
                su3 = su3 + VR(K+(I-1)*Maxi,INDN)
                su2 = su2 + rk(I,JJ,K)*VR(K+(I-1)*Maxi,INDN)
        ENDDO
        write(*,'(I5,2X,I5,2X,ES15.5)') I, JJ, su2/su3
        write(5000+KK,'(ES15.5)') su2/su3
        ENDIF
        ENDDO
        ENDDO

        write(*,*) 
        write(*,*) "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"
        write(*,*)        

        write(*,*) "The first lowest eigenvalue"
        write(*,'(I5,2X,ES15.5)') Jev, WR(Jev)

        J=NNN + 1
        INDM = Jev
        INDN = INDM
        suE(J)=0.0D0
        DO I=1, MM
                suE(J)=suE(J) + VR(I,INDN)
        ENDDO

        DO I=1, MM
                VR(I,INDN) = VR(I,INDN)/suE(J)
        ENDDO
 
        KK = 0
        DO I=1, NN
        DO JJ=1, NN + Np
        IF(I.NE.JJ) THEN
        su2 = 0.0D0
        su3 = 0.0D0
        KK = KK + 1
        DO K=1, Maxi
                INDN = Jev
                su3 = su3 + VR(K+(I-1)*Maxi,INDN)
                su2 = su2 + rk(I,JJ,K)*VR(K+(I-1)*Maxi,INDN)
        ENDDO
        write(*,'(I5,2X,I5,2X,ES15.5)') I, JJ, su2/su3
        write(4000+KK,'(ES15.5)') su2/su3
        ENDIF
        ENDDO
        ENDDO

        write(*,*)
        write(*,*) "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"
        write(*,*)
	
9993     FORMAT(I5,2X,(ES20.10E3,2X))
9994     FORMAT((ES20.10E3,2X))
9995     FORMAT(2(I5,2X),(ES20.10E3,2X))
9996     FORMAT (ES20.10E3,2X,ES20.10E3)

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


        DEALLOCATE ( A, BA, FEJ )
        DEALLOCATE ( rktp, rk, SS, DS )
        DEALLOCATE ( SP, SM )

1111        RETURN
	END

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

