!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CCC
!CCC Name       : SUBROUTINE CalFC 
!CCC Purpose    : To compute a collision frequency (1/s) at a pair (T, P) 
!CCC 
!CCC Authors    : Thanh Lam Nguyen (thanhlamnguyen08@gmail.com)
!CCC              John F. Stanton  (jfstanton@gmail.com)
!CCC
!CCC Date       : Apr. 18, 2014
!CCC Where      : The University of Texas at Austin
!CCC
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        SUBROUTINE CalFC(CF,T,P,m1,Dc1,xma1,m2,Dc2,xma2,MODEL)

        IMPLICIT NONE

integer, parameter :: qp = selected_real_kind(15, 307)

!!!  CF         : collison frequency (1/s)
!!!  T          : temperature (K)
!!!  P          : pressure (atm)
!!!  m1, m2     : masses of bath gas and intermediates/nascent adducts
!!!                     (g/mol)
!!!  Dc1, xma1  : Lennard-Jones collision parameters for bath gas: Sigma
!!!                     (Angstrom) and Eps/kB (K)
!!!  Dc2, xma2  : Lennard-Jones collision parameters for
!!!                     intermediates/nascent adducts


real(kind=qp) :: CF,T,P,m1,Dc1,xma1,m2,Dc2,xma2
real(kind=qp) :: Dc, xma
real(kind=qp) :: PI, No, RT, mo
        Parameter (No = 6.02214D+23, RT=8.31451D+0)
real(kind=qp) :: tp, tp1, tp2, tp3, Z, Po

        Character*4 MODEL


!!! T : temperature (K)
!!! P : pressure (atm)

!!! No : Avogado constant (in 1/mole)
!!! RT : gas constant (in J/mol/K)

        PI = acos(-1.0D0)

        Dc = (Dc1 + Dc2)/2
        xma = sqrt(xma1*xma2)

        IF(MODEL.EQ.'TROE') THEN
                tp = 1.0D0 / ( 0.636D0 + 0.246D0*log(T/xma) )
        ELSE IF(MODEL.EQ.'RPS') THEN   !!! Reid-Prausnitz-Sherwood, 1977
                tp = 1.16145D0/((T/xma)**0.14874D0)
                tp = tp + 0.52487D0*exp(-0.7732D0*T/xma)
                tp = tp + 2.16178D0*exp(-2.43787D0*T/xma)
        ELSE
                tp = 1.16145D0/((T/xma)**0.14874D0)
                tp = tp + 0.52487D0*exp(-0.7732D0*T/xma)
                tp = tp + 2.16178D0*exp(-2.43787D0*T/xma)
        ENDIF

!!!
!!!  tp : collision integral
!!!

        Dc = Dc*1.0D-8
        mo = m1*m2/(m1+m2)/1.0D+3  !!! kg/mol
        tp1 = (RT*T/mo)*1.0D+4
        tp2 = sqrt(8*PI*tp1)
        Z = Dc*tp2*Dc   !!! Z-Lennard-Jones (cm3/molecule/s)
        tp3 = (8.20578D-2)*T*(1.D+3)/No !cm3*atm/molecules
        Po = P/tp3   !!! [M]-molar concentration (molecules/cm3)
        CF = Z*Po    !!!  Z[M] (in Hz)

        CF = CF*tp   !!! collision frequency

        END



